/*
 *  analyzer -- PHOBOS raw-data analyzer main routine
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <sys/socket.h>
#include <pwd.h>
#include "crashm.h"
#include "analyzer.h"

static unsigned short buffer[ACC_BLK_SIZE];
static unsigned int   qwdbuf[ACC_BLK_SIZE];
static unsigned int   scaler[SCL_NUM_DATA];
static unsigned int   nblock=0, mblock=0, lblock=0,
                      nevent=0, mevent=0, levent=0, analyzed=0;

static int  interrupt=0;
static int  intreturn=0;
static void sighandler()
{
  interrupt = 1;
  intreturn = usrast_();
  signal(SIGINT, sighandler);
}

static int  tenblock=0;

static int  summary_out=0;
static void summary()
{
  if (intreturn) puts("\n** stopped by interrupt. **");
  nblock = mblock + lblock;
  printf("\nanalysis summary:\n");
  printf("  %-8s%12s%12s\n", " "       , "blocks", "events");
  printf("  %-8s%12d%12d\n", "total"   , nblock  , nevent);
  printf("  %-8s%12d%12d\n", "valid"   , mblock  , mevent);
  printf("  %-8s%12d%12d\n", "illigal" , lblock  , levent);
  printf("  %-8s%12s%12d\n", "analyzed", " "     , analyzed);
  puts  ("  * comment blocks are not included.\n"); fflush(stdout);
  summary_out = 1;
}
static int blockman()
{
  char  runno[9], comment[81];
  int   i, seqnum=1, buflen=ACC_BLK_LEN(buffer)/*in word*/;
  if (tenblock && nblock%10 == 0)
    {fprintf(stderr, " %d-block\r", nblock); fflush(stderr);}
  switch (ACC_BLK_ID(buffer)) {
  case ACC_BID_STA:
    nblock = mblock = lblock =
    nevent = mevent = levent = analyzed = summary_out = 0;
    bzero(scaler, sizeof(scaler));
    strncpy(runno  , (char *)buffer+BLK_RUN_OFS,  8); runno[8] = 0;
    strncpy(comment, (char *)buffer+BLK_CMT_OFS, 80); comment[80] = 0;
    hststa_(runno, comment, 8, 80);
    printf("\nStart %s\n%s\n", runno, comment); fflush(stdout);
    return intreturn;
  case ACC_BID_STO:
    hststo_(scaler);
    strncpy(runno  , (char *)buffer+BLK_RUN_OFS,  8); runno[8] = 0;
    strncpy(comment, (char *)buffer+BLK_CMT_OFS, 80); comment[80] = 0;
    printf("\nStop  %s\n%s\n", runno, comment); fflush(stdout);
    summary();
    return intreturn;
  case ACC_BID_RAW:
    mblock++;
    nevent += ACC_BLK_EVC(buffer);
    for (i = ACC_HDR_LEN; i < buflen; ) {
      int j=0, event_len=buffer[i+1]/sizeof(short);
      if (intreturn) goto abort;
      if ((buffer[i] != ACC_EVT_HDR(seqnum)) ||
	  (event_len < 2 /* length of EID & EVTLEN */)) {
	levent++;
	printf("Illigal header %4.4x for event# %d at word %d in block %d\n",
	       buffer[i], seqnum, i, mblock+lblock);
	if (++seqnum > ACC_BLK_EVC(buffer)) break;
	do {if (++i >= buflen) goto abort;}
	while (buffer[i] != ACC_EVT_HDR(seqnum));
	continue;}
      mevent++;
      do qwdbuf[j++] = buffer[i++]; while (j < event_len);
      switch (event1_((int *)qwdbuf, &event_len)) {
      case 0: analyzed++; break;
      case 1:             break;
      case 2: goto abort;
      case 3: return 1;
      }
      if (++seqnum > ACC_BLK_EVC(buffer)) break;
    } /* for (i = ACC_HDR_LEN; i < buflen; ) */
  abort:
//#define MONITOR
#ifdef MONITOR
    printf("mblock=%d  lblock=%d  nevent=%d  mevent=%d  levent=%d\n",
	   mblock, lblock, nevent, mevent, levent); fflush(stdout);
#endif
    for (i = 0; i < SCL_NUM_DATA; i++)
      scaler[i] += ((unsigned int *)&buffer[SCL_BLK_OFS])[i];
    return (endblk_((unsigned int *)&buffer[SCL_BLK_OFS]) || intreturn);
  case ACC_BID_BAD:
  default:
    lblock++;
    return intreturn;
  }
}

static char  user[20]=USERNAME;
static char  host[80]=HOSTNAME;
static char  sampler[256]=SAMPLER;
static char  ID[256]="\0";

int main(int argc, char *argv[])
{
  char   c, localhost[80], pwd[20], s[256], *proc;
  extern char *optarg;
  extern int   optind, opterr, optopt;
  FILE  *fd;
  int    sd=0, i;
  if ((proc = rindex(argv[0], '/')) != NULL) proc++; else proc = argv[0];
  while ((c = getopt(argc, argv, "l:r:s:bh")) != -1)
    switch(c) {
    case 'l':
      strcpy(user, optarg);
      break;
    case 'r':
      sprintf(s, "nslookup %s 2> /dev/null | awk '/Name:/{print $2}'", optarg);
      fd = popen(s, "r"); s[0] = 0; fgets(s, sizeof(s), fd); fclose(fd);
      if (!s[0]) {fprintf(stderr, "illegal host -- %s\n", optarg); exit(2);}
      s[strlen(s)-1] = 0 /* to eliminate "\n" */; strcpy(host, s);
      break;
    case 's':
      strcpy(sampler, optarg);
      break;
    case 'b':
      tenblock = 1;
      break;
    case 'h':
    default :
      printf("Usage:  %s [-l user] [-r host] [-s sampler] [-b] [-h] [id_name]
", proc);
      exit(2);
    }
  if (optind < argc) strncpy(ID, argv[optind], sizeof(ID));
  if (!ID[0]) strcpy(ID, proc);
  fd = popen("hostname --long", "r");
  fgets(localhost, sizeof(localhost), fd);
  pclose(fd);
  localhost[strlen(localhost)-1] = 0;
  if (host[0] == 0 || !strcmp(host, localhost)) { /* initiate local sampler */
    /* discard SIGINT at this stage, so as not to pass SIGINT to sub-proc. */
    signal(SIGINT, SIG_IGN); 
    sprintf(s, sampler, ID);
    if ((fd = popen(s, "r")) == NULL) {fputs("popen error" ,stderr); exit(2);}
  } else {                                       /* initiate remote sampler */
    struct servent *serv;
    char  *rhost = (char *)host;
    if (isatty(0)) {       /* invoked from terminal */
      sprintf(s, "Password (%s@%s): ", user, host);
      strcpy(pwd, getpass(s));
    } else {               /* invoked as subprocess */
      printf("Password (%s@%s): ", user, host);
      gets(pwd);
    }
    if ((serv = getservbyname("exec", "tcp")) == NULL)
      {fputs("exec/tcp: service not available", stderr); exit(2);}
    sprintf(s, sampler, ID);
    if ((sd = rexec((char **)&rhost, serv->s_port, user,
		    pwd, s, (int *)NULL)) == -1)
      {fputs("rexec fail", stderr); exit(2);}
    fd = fdopen(sd, "r");
  }
  setvbuf(fd, (char *)NULL, _IOFBF, ACC_BLK_SIZE*sizeof(short));
  signal(SIGINT, sighandler);
  i = SCL_NUM_DATA; 
  hstini_(ID, &i, strlen(ID));
  while (1) {
    if (fread(buffer, sizeof(short), ACC_BLK_SIZE, fd) == ACC_BLK_SIZE)
      {if (blockman()) break;}
    else {
      if (feof(fd)) break;
      usleep(10000L);              /* trap for SysV-style signal processing, */
      if (interrupt) interrupt = 0;/* which aborts system calls at signal. */
      else {if (*(char *)buffer == '#') fputs((char *)buffer, stderr); break;}
    }
  }
  if (!strcmp(host, localhost)) pclose(fd);
  else {fclose(fd); shutdown(sd, 2); close(sd);}
  if (!summary_out) summary();
  hstend_(scaler);
  exit(0);
}

/* dummy MAIN routine for libf2c.so */

void MAIN__(){}
