/*
 *  recorder -- PHOBOS raw-data recorder
 */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include "camac.h"
#include "k3922_reg.h"
#include "acc_conf.h"
#include "crashm.h"

int rmt_open();
int rmt_close();

struct naf_data cc;
#define ccnaf(N,A,F,WSIZE) \
cc.mode = SINGLE | WSIZE;\
cc.naf  = NAFGEN(N,A,F);\
ioctl(fd, CCIOC_SINGLE, &cc);
#define ccnafw(N,A,F,WSIZE,D) \
cc.data = D;\
ccnaf(N,A,F,WSIZE);
#define ccnafq(N,A,F,WSIZE) \
cc.mode = SINGLE | WSIZE;\
cc.naf  = NAFGEN(N,A,F);\
{int i=0;\
do ioctl(fd, CCIOC_SINGLE, &cc);\
while ((cc.xq & (NOX | NOQ)) && (i++ < 20));}
#define ccnafqw(N,A,F,WSIZE,D) \
cc.data = D;\
ccnafq(N,A,F,WSIZE);

int     shmid;
CRASHM *shmp;
int     semid, sem_buffer, sem_sample;

int main(int argc, char *argv[])
{
  int           fd, wp, *p, i;
  struct sembuf semcmd;
  unsigned int  scaler[SCL_NUM_DATA];

  setvbuf(stdout, (char *)NULL, _IONBF, 0); /* set stdout unbuffered */
  if ((shmid = shmget(SHMKEY, sizeof(CRASHM), IPC_CREAT|0600)) == -1)
    {puts("! Cannot create shared memory"); exit(2);}
  if ((semid = semget(SEMKEY, SEM_NUM       , IPC_CREAT|0600)) == -1 ||
      (sem_buffer = semget(SEM_BUFFER, CRA_BLK_NUM, IPC_CREAT|0600)) == -1 ||
      (sem_sample = semget(SEM_SAMPLE, CRA_BLK_NUM, IPC_CREAT|0600)) == -1)
    {puts("! Cannot create semaphore"); exit(2);}
  shmp = (CRASHM *)shmat(shmid, 0, 0);
  if ((fd = open("/dev/cc0", O_RDWR)) == -1)
    {puts("! CAMAC device open error."); exit(2);}

  while (!ioctl(fd, CCIOC_WAIT_LAM, 0)) { /* wait for LAM */
    if (shmp->acquire == -1) break;
    if (shmp->acquire) shmp->wp++; else ioctl(fd, CCIOC_DISABLE_INT, 0);
    wp = shmp->wp % CRA_BLK_NUM;
    sem_op(sem_sample, semcmd, wp, 0); /* wait for sampler(s) done */
    sem_op(sem_buffer, semcmd, wp, 1); /* lock buffer */

    /* get scalers (serves to reset scalers for start comment) */
    ioctl(fd, CCIOC_SET_CRATE, SCL_CRATE);
    cc.mode = QSCAN | WSIZE24;
    cc.naf  = NAFGEN(SCL_N,0,SCL_F);
    ioctl(fd, CCIOC_SETUP_DMA, &cc);
    if ((i = (read(fd, scaler, SCL_NUM_DATA*sizeof(int))
	      != SCL_NUM_DATA*sizeof(int))))
      printf("DMA error for scalers (%d)\n", shmp->wp);
    ioctl(fd, CCIOC_SET_CRATE, ACC_CRATE);

    if (shmp->acquire) {
      int bcbf;
      /* get a buffer with DMA */
      ccnafq (ACC_N,0,10,WSIZE24);		/* clear ACC LAM */
      ccnafqw(ACC_N,0,17,WSIZE24,ACC_STR);	/* set MAR to STR */
      ccnafq (ACC_N,0, 0,WSIZE16);		/* get STR */
      if (cc.data & ACC_STR_BERR || !(cc.data & ACC_STR_DMA))
	{printf("STR error (%#x)\n", cc.data);  /* and do something ? */}
      ccnafqw(ACC_N,0,17,WSIZE24,ACC_VIR);	/* set MAR to VIR */
      ccnafq (ACC_N,0, 0,WSIZE16);		/* get VIR */
      if (cc.data & ACC_VIR_ERR || !(cc.data & (ACC_VIR_BUF1|ACC_VIR_BUF2)))
	{printf("VIR error (%#x)\n", cc.data);  /* and do something ? */}
      ccnafqw(ACC_N,0,17,WSIZE24,ACC_BPT);	/* set MAR to BPT */
      ccnafq (ACC_N,1, 0,WSIZE24);		/* get BPT */
      bcbf = ACC_BCBF(cc.data);
      ccnafqw(ACC_N,0,17,WSIZE24,bcbf);		/* set MAR to BCBF */
      ccnafq (ACC_N,1, 0,WSIZE24);		/* get BCBF */
      ccnafqw(ACC_N,0,17,WSIZE24,cc.data);	/* set MAR to BCBA */
      ccnafq (ACC_N,1, 0,WSIZE24);		/* get BCBA */
      ccnafqw(ACC_N,0,17,WSIZE24,cc.data);	/* set MAR to buff. adr. */
#ifdef ACC_DOUBLE_BUFFER
      /* output pulse to outreg notifying end of interrupt */
      ccnafqw(ACC_NOPR,0,17,WSIZE16,ACC_ENDISR);
#endif
      cc.mode = QREPEAT | WSIZE16;
      cc.naf  = NAFGEN(ACC_N,0,0);
      ioctl(fd, CCIOC_SETUP_DMA, &cc);
      if (read(fd, shmp->buffer[wp], ACC_BLK_SIZE*sizeof(short))
	  != ACC_BLK_SIZE*sizeof(short))
	ACC_BLK_ID(shmp->buffer[wp]) = ACC_BID_BAD;
      /* clear BCBF */
      ccnafqw(ACC_N,0,17,WSIZE24,bcbf);		/* set MAR to BCBF */
      ccnafqw(ACC_N,0,16,WSIZE16,0);		/* clear BCBF */
#ifndef ACC_DOUBLE_BUFFER
      /* output pulse to outreg notifying end of interrupt */
      ccnafw (ACC_NOPR,0,17,WSIZE16,ACC_ENDISR);
#endif
      /* mark bad block-ID if scaler error or data overrun */
      if (i || (ACC_BLK_LEN(shmp->buffer[wp]) > SCL_BLK_OFS*sizeof(short)))
	ACC_BLK_ID(shmp->buffer[wp]) = ACC_BID_BAD;
      /* copy scalers and add them to "summed" scaler */
      p = (int *)&(shmp->buffer[wp][SCL_BLK_OFS]);
      for (i = 0; i < SCL_NUM_DATA; i++)
	{*p = scaler[i]; shmp->scaler[i] += *(p++);}
      if (ACC_BLK_ID(shmp->buffer[wp]) == ACC_BID_RAW)
	shmp->nblock++; else shmp->badblock++;

    } else {
      /* write comment block */
      bzero(shmp->buffer[wp], ACC_BLK_SIZE*sizeof(short));
      ACC_BLK_ID(shmp->buffer[wp]) =
	(shmp->sto_time[0] ? ACC_BID_STO : ACC_BID_STA);
      sprintf((char *)(&shmp->buffer[wp])+BLK_RUN_OFS, 
	      "RUN-%4.4d  START => %8s  STOP => %8s   Print => %8s  %9s", 
	      shmp->runno, shmp->sta_time, shmp->sto_time, shmp->pri_time, 
	      shmp->sta_date);
      sprintf((char *)(&shmp->buffer[wp])+BLK_CMT_OFS,
	      "%-160s", shmp->comment);
      usleep(50000L);
    }

    sem_op(sem_buffer, semcmd, wp, -1); /* free buffer */
    if ((i = semctl(semid, sem_newdat, GETNCNT, 0)))
      sem_op(semid, semcmd, sem_newdat, i); /* wake up sampler(s) if waiting */
//#define MONITOR
#ifdef MONITOR
    printf("wp=%4d, %4.4x\n",
	   shmp->wp, shmp->buffer[wp][ACC_BLK_SIZE-2]);
    for (i = 0; i < 64; i++) printf(" %4.4x", shmp->buffer[wp][i]); puts("");
    p = (int *)&(shmp->buffer[wp][SCL_BLK_OFS]);
    for (i = 0; i < SCL_NUM_DATA; i++) printf("%10d",*(p++));
#endif
    if (shmp->path[0]) {                /* write data to storage */
      if (ACC_BLK_ID(shmp->buffer[wp]) == ACC_BID_STA) if (rmt_open()) break;
      if (write(shmp->fd, shmp->buffer[wp], ACC_BLK_SIZE*sizeof(short))
	  != ACC_BLK_SIZE*sizeof(short)) {
	perror("write"); puts("# write error");
	close(shmp->fd); shmp->path[0] = 0;}
      if (shmp->mtape) shmp->mtblock++;
      if (ACC_BLK_ID(shmp->buffer[wp]) == ACC_BID_STO) if (rmt_close()) break;
    }
    semctl(semid, sem_record, SETVAL, 0); /* wake up cmmnder if waiting */
    ioctl(fd, CCIOC_ENABLE_INT, 0);
  }

  close(fd);
  shmdt((char *)shmp);
  semctl(sem_buffer, 0, IPC_RMID, 0);
  semctl(sem_sample, 0, IPC_RMID, 0);
  exit(0);
}

#include <netdb.h>
#include <sys/socket.h>

struct  servent *serv;

int rmt_open()
{
  char  *host=(char *)shmp->host, s[256];
  if (shmp->remote) serv = getservbyname("exec", "tcp");
  if (shmp->mtape) {    /* record to tape */
    if (!shmp->remote) {
      if ((shmp->fd = open(shmp->path, O_CREAT | O_WRONLY, 0644)) == -1)
	{puts("# mt open error."); shmp->fd = 0; return -1;}
    } else {
      sprintf(s, "/bin/dd bs=%d of=%s",
	      ACC_BLK_SIZE*sizeof(short), shmp->path);
      if ((shmp->fd = rexec((char **)(&host), serv->s_port, shmp->user,
			    shmp->pwd, s, (int *)NULL)) == -1)
	{puts("# remote mt open error."); shmp->fd = 0; return -1;}
    }
  } else {              /* record to disk */
    if (!shmp->remote) {
      sprintf(s, "%s/run%4.4d.dat", shmp->path, shmp->runno);
      if ((shmp->fd = open(s, O_CREAT | O_WRONLY, 0644)) == -1)
	{puts("# file open error."); shmp->fd = 0; return -1;}
    } else {
      sprintf(s, "/bin/dd bs=%d of=%s/run%4.4d.dat", 
	      ACC_BLK_SIZE*sizeof(short), shmp->path, shmp->runno);
      if ((shmp->fd = rexec((char **)(&host), serv->s_port, shmp->user,
			    shmp->pwd, s, (int *)NULL)) == -1)
	{puts("# remote file open error."); shmp->fd = 0; return -1;}
    }
  }
  return 0;
}

int rmt_close()
{
  char  *host=(char *)shmp->host, s[80];
  int    sd;
  FILE  *fd;
  if (shmp->remote) shutdown(shmp->fd, 2);
  close(shmp->fd);
  shmp->fd = shmp->path[0] = 0;
  printf("<< CLOSE >>\n%d blks (%d bad-blks",
	 shmp->nblock+2, shmp->badblock);
  if (shmp->mtape) {
    int  b;
    sprintf(s, "mt weof 2 -f %s ; mt bsfm 1 -f %s \n", shmp->path, shmp->path);
    if (!shmp->remote)
      return system(s);
    else {
      if ((sd = rexec((char **)(&host), serv->s_port, shmp->user,
		      shmp->pwd, "/bin/sh", (int *)NULL)) == -1)
	{puts("# rexec fail"); return -1;}
      fd = fdopen(sd, "r+");
      fputs(s, fd); fflush(fd); fclose(fd);
      shutdown(sd, 2); close(sd);
    }
    b = shmp->mtblock*ACC_BLK_SIZE*sizeof(short);
    printf(", %d.%d MB on tape", b/1048576, (b%1048576)/104858);
  }
  puts(")");
  return 0;
}
